%{

let make_pterm q e = 
  let rec f l = 
      match l with ((x,y)::[]) -> "pterm1 (" ^ x ^ "\\ ( " ^ y ^ ", " ^ e ^ "))"
      | ((x,y)::xs) -> "pterm2( " ^ x ^ "\\ ( " ^ y ^ ", " ^ (f xs) ^ "))"
  in
    f q

%}


%token <string> LCID
%token <string> UCID
%token <string> TESTNAME
%token LAMBDA COLON DOT STYPE LSQUARE RSQUARE EQUALS
%token LPAREN RPAREN TOP ARROW FORALL LCURLY RCURLY QUOTE
%token EOF COMMA LET IN SEMICOLON DUMMY TURNSTILE STEP QUESTION STAR TEST

%nonassoc DUMMY 
%right DOT ARROW LSQUARE
%nonassoc IN

%start main
%type <string> main

%%

main: 
  TEST QUOTE ids QUOTE EQUALS test { $6 }
| term { "( " ^ $1 ^ " )" }
| typ  { "( " ^ $1 ^ " )" } 
;

ids: 
  UCID { }
| LCID {} 
| LCID ids {}
| UCID ids {}
;

test: 
  context TURNSTILE term COLON typ { "typeof " ^ $1 ^ " (" ^ $3 ^ ") (" ^ $5 ^ ")." }
| term STEP term { "(" ^ $1 ^ ") ---> (" ^ $3 ^ ")." }
| term STEP QUESTION { "(" ^ $1 ^ ") ---> E." }
| term STEP STAR term { "(" ^ $1 ^ ") ===> (" ^ $4 ^ ")." }
;

context: 
  LCURLY RCURLY { "[]" }
| LCURLY bindings RCURLY { "[" ^ $2 ^ "]" }

bindings: 
  LCID COLON typ { "(id_pair " ^ $1 ^ " (" ^ $3 ^ "))" }
| UCID STYPE typ { "(tid_pair " ^ $1 ^ " (" ^ $3 ^ "))" }
| LCID COLON typ COMMA bindings { "(id_pair " ^ $1 ^ " (" ^ $3 ^ ")), " ^ $5 }
| UCID STYPE typ COMMA bindings { "(tid_pair " ^ $1 ^ " (" ^ $3 ^ ")), " ^ $5 }
;

term: 
  LAMBDA LCID COLON typ DOT term { "lam (" ^ $2 ^ "\\ (" ^ $4 ^ ", " ^ $6 ^ "))" }
| LAMBDA UCID STYPE typ DOT term { "tlam (" ^ (String.lowercase $2) ^ "_\\ (" ^ $4 ^ ", " ^ $6 ^ "))" }
| LCURLY labeltermseq RCURLY { "rcd [" ^ $2 ^ " ]" }
| LET pattern EQUALS term IN term { let pattern,pterm' =  $2 in
                                    let pterm = make_pterm pterm' $6 in
                                     "patt (" ^ $4 ^ ") (" ^ pattern ^ ") (" ^ pterm ^ ")" }
| appterm { $1 } 
;

appterm: 
  pathterm  { $1 }
| appterm LSQUARE typ RSQUARE { "tapp (" ^ $1 ^ ") (" ^ $3 ^ ")" }
| appterm pathterm { "app (" ^ $1 ^ ") (" ^ $2 ^ ")" }
;

pathterm: 
  pathterm DOT LCID { "proj (" ^ $1 ^ ") \"" ^ $3 ^ "\"" }
| aterm { $1 }
;

aterm: 
  LCID { "var " ^ $1 }
| LPAREN term RPAREN { "(" ^ $2 ^ ")" }
;

typ: 
  UCID { "tvar " ^ (String.lowercase $1) ^ "_" }
| TOP  { "top" }
| typ ARROW typ { $1 ^ " ==> " ^ $3 } 
| FORALL UCID STYPE typ DOT typ { "tall (" ^ (String.lowercase $2) ^ "_\\(" ^ $4 ^ ", " ^ $6 ^ "))" }
| LCURLY labeltypseq RCURLY { "trcd [" ^ $2 ^ "]" }
| LPAREN typ RPAREN { "( " ^ $2 ^ " )" }
;

labeltermseq: 
  LCID EQUALS term   { "(\"" ^ $1 ^ "\", " ^ $3 ^ ")" }
| LCID EQUALS term COMMA labeltermseq { "(\"" ^ $1 ^ "\", " ^ $3 ^ "), " ^ $5 }
;

labeltypseq: 
  LCID EQUALS typ   { "(\"" ^ $1 ^ "\", " ^ $3 ^ ")" }
| LCID EQUALS typ COMMA labeltypseq { "(\"" ^ $1 ^ ", " ^ $3 ^ "), " ^ $5 }
;

pattern: 
  LCID COLON typ   { "pvar (" ^ $3 ^ ")", [($1, $3)] }
| LCURLY labelpatternseq RCURLY { let patt,pterm = $2 in 
                                    "prcd [" ^ patt ^ "]", pterm }
;

labelpatternseq: 
  LCID EQUALS pattern { let patt,pterm = $3 in "(\"" ^ $1 ^ "\", " ^ patt ^ ")", pterm }
| LCID EQUALS pattern COMMA labelpatternseq { let patt,pterm = $3 in 
                                              let patt',pterm' = $5 in 
					      "(\"" ^ $1 ^ "\", " ^ patt ^ "), " ^ patt', pterm@pterm' }
;
